import 'dart:ffi';
import 'package:csounddart/CsoundDartUtilities.dart';
import 'package:csounddart/csoundAndroid.dart';
import 'package:ffi/ffi.dart';
import 'dart:io';
import 'dart:isolate';
import 'csound.dart';
import"audio_utilities.dart";

//import 'package:permission_handler/permission_handler.dart';

import 'csound_path_finder.dart';


Future<void> requestPermissions() async
{
  /*
  if(!Platform.isAndroid) return;
   Map<Permission, PermissionStatus> lst = await [Permission.storage,
     Permission.microphone, Permission.mediaLibrary, Permission.sensors].request();
   */
}


DynamicLibrary loadLib()
{
  if(Platform.isAndroid) {
    return DynamicLibrary.process();
  }
  String libraryPath = CsoundPathFinder.findPath();
  DynamicLibrary lib =  DynamicLibrary.open(libraryPath);
  print("Library loaded --> $libraryPath" );
  return lib;
}


final lib  = loadLib();

//Function signatures
// names are not really significant

typedef voidFunc = Void Function();

typedef destroy_func = Void Function(Pointer<Void> csound);
typedef destroyFunc = void Function(Pointer<Void> csound);

typedef get_info_func = Double Function(Pointer<Void> csound);
typedef getInfoFunc = double Function(Pointer<Void> csound);

typedef get_uint_func = Int32 Function(Pointer<Void> csound);
typedef getUintFunc = int Function(Pointer<Void> csound);

typedef init_func = Int32 Function(Int32 flags);
typedef initFunc = int Function(int flags);

typedef create_func = Pointer<Void> Function(Pointer<Void> hostData);
typedef createFunc = Pointer<Void> Function(Pointer<Void> hostData);

typedef compile_csd_text_func = Int32 Function(Pointer<Void> csound, Pointer<Utf8> csdString);
typedef compileCsdTextFunc = int Function(Pointer<Void> csound, Pointer<Utf8> csdString);

typedef eval_code_func = Double Function(Pointer<Void> csound, Pointer<Utf8> string);
typedef evalCodeFunc = double Function(Pointer<Void> csound, Pointer<Utf8> string);

typedef perform_func = Int32 Function(Pointer<Void> csound);
typedef performFunc = int Function(Pointer<Void> csound);

typedef set_control_func = Void Function(Pointer<Void> csound, Pointer<Utf8> channelName, Double value);
typedef setControlFunc = void Function(Pointer<Void> csound, Pointer<Utf8> channelName, double value);

typedef get_control_func = Double Function(Pointer<Void> csound, Pointer<Utf8> channelName, Pointer<Int32> err);
typedef getControlFunc = double Function(Pointer<Void> csound, Pointer<Utf8> channelName, Pointer<Int32> err);

typedef set_string_func = Void Function(Pointer<Void> csound, Pointer<Utf8> channelName, Pointer<Utf8> value);
typedef setStringFunc = void Function(Pointer<Void> csound, Pointer<Utf8> channelName, Pointer<Utf8> value);

typedef perf_routine_func = Int32 Function(Pointer<Void> data);
typedef create_thread_func = Pointer<Void> Function(Pointer<NativeFunction<perf_routine_func>> perf, Pointer<Void> data);

typedef get_table_func = Int32 Function(Pointer<Void> csound, Pointer<Pointer<Double>> tablePtr, Int32 tableNum);
typedef getTableFunc = int Function(Pointer<Void> csound, Pointer<Pointer<Double>> tablePtr, int tableNum);

typedef get_audio_channel_func = Void Function(Pointer<Void> csound, Pointer<Utf8> channelName, Pointer<Double> value);
typedef getAudioChannelFunc = void Function(Pointer<Void> csound, Pointer<Utf8> channelName, Pointer<Double> value);


typedef intTriggerFunc = Int32 Function(Pointer<Void> csoundInstance);



// Functions wrapping.

final csoundInitializePointer = lib.lookup<NativeFunction<init_func>>("csoundInitialize");
final csoundInitialize = csoundInitializePointer.asFunction<initFunc>();

final csoundCreatePointer = lib.lookup<NativeFunction<create_func>>("csoundCreate");
final csoundCreate = csoundCreatePointer.asFunction<createFunc>();

final csoundCompileCsdTextPointer = lib.lookup<NativeFunction<compile_csd_text_func>>("csoundCompileCsdText");
final csoundCompileCsdText = csoundCompileCsdTextPointer.asFunction<compileCsdTextFunc>();

final csoundPerformPointer = lib.lookup<NativeFunction<perform_func>>("csoundPerform");
final csoundPerform = csoundPerformPointer.asFunction<performFunc>();

final csoundPerformKsmpsPointer = lib.lookup<NativeFunction<perform_func>>("csoundPerformKsmps");
final csoundPerformKsmps = csoundPerformKsmpsPointer.asFunction<performFunc>();

final csoundStartPointer = lib.lookup<NativeFunction<perform_func>>("csoundStart");
final csoundStart = csoundStartPointer.asFunction<performFunc>();

final csoundDestroyPointer = lib.lookup<NativeFunction<destroy_func>>("csoundDestroy");
final csoundDestroy = csoundDestroyPointer.asFunction<destroyFunc>();

final csoundCompileOrcPointer = lib.lookup<NativeFunction<compile_csd_text_func>>("csoundCompileOrc");
final csoundCompileOrc = csoundCompileOrcPointer.asFunction<compileCsdTextFunc>();

final csoundReadScorePointer = lib.lookup<NativeFunction<compile_csd_text_func>>("csoundReadScore");
final csoundReadScore = csoundReadScorePointer.asFunction<compileCsdTextFunc>();

final csoundEvalCodePointer = lib.lookup<NativeFunction<eval_code_func>>("csoundEvalCode");
final csoundEvalCode = csoundEvalCodePointer.asFunction<evalCodeFunc>();

final csoundCompilePointer = lib.lookup<NativeFunction<compile_csd_text_func>>("csoundCompile");
final csoundCompile = csoundCompilePointer.asFunction<compileCsdTextFunc>();

final csoundStopPointer = lib.lookup<NativeFunction<destroy_func>>("csoundStop");
final csoundStop = csoundStopPointer.asFunction<destroyFunc>();

final csoundCleanupPointer = lib.lookup<NativeFunction<destroy_func>>("csoundStop");
final csoundCleanup = csoundCleanupPointer.asFunction<destroyFunc>();

final csoundResetPointer = lib.lookup<NativeFunction<destroy_func>>("csoundReset");
final csoundReset = csoundResetPointer.asFunction<destroyFunc>();

final csoundGetSrPointer = lib.lookup<NativeFunction<get_info_func>>("csoundGetSr");
final csoundGetSr = csoundGetSrPointer.asFunction<getInfoFunc>();

final csoundGet0dBFSPointer = lib.lookup<NativeFunction<get_info_func>>("csoundGet0dBFS");
final csoundGet0dBFS = csoundGet0dBFSPointer.asFunction<getInfoFunc>();

final csoundGetKsmpsPointer = lib.lookup<NativeFunction<get_uint_func>>("csoundGetKsmps");
final csoundGetKsmps = csoundGetKsmpsPointer.asFunction<getUintFunc>();

final csoundGetNchnlsPointer = lib.lookup<NativeFunction<get_uint_func>>("csoundGetNchnls");
final csoundGetNchnls = csoundGetNchnlsPointer.asFunction<getUintFunc>();

final csoundGetNchnlsIsInputPointer = lib.lookup<NativeFunction<get_uint_func>>("csoundGetNchnlsIsInput");
final csoundGetNchnlsIsInput = csoundGetNchnlsIsInputPointer.asFunction<getUintFunc>();

final csoundSetOptionPointer = lib.lookup<NativeFunction<compile_csd_text_func>>("csoundSetOption");
final csoundSetOption = csoundSetOptionPointer.asFunction<compileCsdTextFunc>();


final csoundSetControlChannelPointer = 
lib.lookup<NativeFunction<set_control_func>>("csoundSetControlChannel");
final csoundSetControlChannel = csoundSetControlChannelPointer.asFunction<setControlFunc>();

final csoundGetControlChannelPointer = 
lib.lookup<NativeFunction<get_control_func>>("csoundGetControlChannel");
final csoundGetControlChannel = csoundGetControlChannelPointer.asFunction<getControlFunc>();

final csoundSetStringChannelPointer = 
lib.lookup<NativeFunction<set_string_func>>("csoundSetStringChannel");
final csoundSetStringChannel = csoundSetStringChannelPointer.asFunction<setStringFunc>();

final csoundGetStringChannelPointer = 
lib.lookup<NativeFunction<set_string_func>>("csoundGetStringChannel");
final csoundGetStringChannelC = csoundGetStringChannelPointer.asFunction<setStringFunc>();

final csoundCreateThreadPointer = 
lib.lookup<NativeFunction<create_thread_func>>("csoundCreateThread");
final csoundCreateThread = csoundCreateThreadPointer.asFunction<create_thread_func>();

final csoundGetTablePointer = lib.lookup<NativeFunction<get_table_func>>("csoundGetTable");
final csoundGetTableC = csoundGetTablePointer.asFunction<getTableFunc>();

final csoundGetAudioChannelPointer = lib.lookup<NativeFunction<get_audio_channel_func>>("csoundGetAudioChannel");
final csoundGetAudioChannel = csoundGetAudioChannelPointer.asFunction<getAudioChannelFunc>();

//wrapper function (string utf8 conversion)
String csoundGetStringChannel(Pointer<Void> csound, String channelName) {
  Pointer<Utf8> ptr = calloc.allocate(128);
  csoundGetStringChannelC(csound, StringUtf8Pointer(channelName).toNativeUtf8(), ptr);
  return Utf8Pointer(ptr).toDartString();
}


List<double> csoundGetTable(Pointer<Void> csound, int table)
{
  Pointer<Pointer<Double>> ptr = Pointer.fromAddress(nullptr.address);
  int res = csoundGetTableC(csound, ptr, table);
  Pointer<Double> myfltPtr = ptr[0];
  List<double> l = List.generate(res, (index) => 0.0);
  for(int i = 0; i < res; i++) l[i] = myfltPtr[i];
  return l;
}

void csoundSetTable(Pointer<Void> csound, List<double> value, int table)
{
  Pointer<Pointer<Double>> ptr = Pointer.fromAddress(nullptr.address);
  int res = csoundGetTableC(csound, ptr, table);
  for(int i = 0; i < res; i++) ptr[0][i] = value[i];
}



// Implementation of CsoundDart class
class MessageTag {
  static const String MESSAGE_TYPE = "1";
  static const String MESSAGE_CONTENT = "2";
  static const String CONTROL_CHANNEL_VALUE = "3";
  static const String STRING_CHANNEL_VALUE = "4";
  static const String CHANNEL_NAME = "5";
  static const String CALLBACK_INDEX = "6";
  static const String TABLE_INDEX = "7";
  static const String AUDIO_CHANNEL_DATA = "8";
  static const String AUDIO_CHANNEL_PTR = "9";

  static const String TYPE_CONTROL = "10";
  static const String TYPE_AUDIO = "11";
}

class MessageType {
  static const String COMPILE_CSD_TEXT = "1";
  static const String COMPILE_CSD = "2";
  static const String COMPILE_ORC = "3";
  static const String READ_SCORE = "4";
  static const String PERFORM = "5";
  static const String STOP =  "6";
  static const String SET_CONTROL_CHANNEL = "7";
  static const String GET_CONTROL_CHANNEL = "8";
  static const String DELETE = "9";
  static const String SET_CONTROL_CHANNEL_CALLBACK = "10";
  
  static const String SET_STRING_CHANNEL = "11";
  static const String GET_STRING_CHANNEL = "12";
  static const String SET_STRING_CHANNEL_CALLBACK = "13";
  static const String SET_TABLE = "14";
  static const String SET_TABLE_CALLBACK = "15";

  static const String SET_AUDIO_CHANNEL_CALLBACK = "16";

  static const String GET_KSMPS = "17";
  static const String GET_SR = "18";
  static const String GET_NCHNLS = "19";
  static const String EVAL_CODE = "20";

  static const String INIT_CALLBACK_PORT = "0";

  static const String CUSTOM_MESSAGE = "-1";
}

Map<String, String> generateMessage(String messageType, {String content = ""} )
{
  Map<String, String> map = new Map< String, String>();
  map[MessageTag.MESSAGE_TYPE] = messageType.toString();
  map[MessageTag.MESSAGE_CONTENT] = content;
  return map;
}

Map<String, String> generateControlMessage(String channelName, double value)
{
  Map<String, String> map = new Map<String, String>();
  map[MessageTag.MESSAGE_TYPE] = MessageType.SET_CONTROL_CHANNEL;
  map[MessageTag.CHANNEL_NAME] = channelName;
  map[MessageTag.CONTROL_CHANNEL_VALUE] = value.toString(); 
  return map;
}

class CsoundNative implements Csound {

  CsoundNative() : _callbackPort = ReceivePort() , _receivePort = ReceivePort()
  {
    print("About to start isolate " ) ;
    _startIsolate();

    print("Isolate started" );
  }

  void _initCallbackPort() async
  {
    _sendPort.send([{MessageTag.MESSAGE_TYPE: MessageType.INIT_CALLBACK_PORT}, _callbackPort.sendPort]);
    // Listen for callback control messages
    _callbackPort.listen((message) {
      if(message[1] == MessageTag.TYPE_CONTROL) {
        _callbackList[message[0] - 1].call(message[2]);
      } else if(message[1] == MessageTag.TYPE_AUDIO) {
        //print("Audio callback");
        ConcurrentBuffer buf = ConcurrentBuffer.fromAddress(message[2]);
        //print("buf is created");
        _callbackList[message[0] - 1].call(buf);
        //print("buf is called");
      }
    });
  }

  void dispose()
  {
    _sendMessage(generateMessage(MessageType.DELETE));
    _isolate.kill(priority: Isolate.immediate);
  }

  int compileCsdText(String text) {
      _sendMessage(generateMessage(MessageType.COMPILE_CSD_TEXT, content: text));
      return 0;
  }

  int compileOrc(String orc)
  {

      return 0;
  }

  int compileCsd(String csd)
  {

      return 0;
  }

  int readScore(String score)
  {

      return 0;
  }

  void setControlChannel(String channelName, double value)
  {
    _sendMessage(generateControlMessage(channelName, value));
  }

  Future<double> getControlChannel(String channelName) async
  {
    Map<String, String> m = {MessageTag.MESSAGE_TYPE: MessageType.GET_CONTROL_CHANNEL, MessageTag.CHANNEL_NAME: channelName};
    ReceivePort rport = new ReceivePort();
    _sendPort.send([m, rport.sendPort]);

    double res = 0.0;

    await for(var msg in rport) {
      print(msg);
      res = double.parse(msg[0][MessageTag.CONTROL_CHANNEL_VALUE]);
      break;
    }
    rport.close();

    return res;
  }

  int perform() 
  {   //start performance
      _sendMessage(generateMessage(MessageType.PERFORM));
      //start callback listening
      return 0;
  }

  void stop()
  {
    _sendMessage(generateMessage(MessageType.STOP));
  }

  void setControlChannelCallback(String channelName, Function callback)
  {
    Map<String,String> map = generateMessage(MessageType.SET_CONTROL_CHANNEL_CALLBACK, content: "");
    map[MessageTag.CHANNEL_NAME] = channelName;
    _callbackList.add(callback);
    map[MessageTag.CALLBACK_INDEX] = _callbackList.length.toString();
    _sendPort.send([map, _callbackPort.sendPort]);
  }


  void setAudioChannelCallback(String channel, Function f)
  {
    print("Setting Audio Channel Callback");
    Map<String, String> map = generateMessage(MessageType.SET_AUDIO_CHANNEL_CALLBACK, content: "");
    map[MessageTag.CHANNEL_NAME] = channel;
    _callbackList.add(f);
    map[MessageTag.CALLBACK_INDEX] = _callbackList.length.toString();
    _sendPort.send([map, _callbackPort.sendPort]);
  }

  void setStringChannel(String channelName, String value) {
    Map<String,String> map = {
      MessageTag.MESSAGE_TYPE: MessageType.SET_STRING_CHANNEL,
      MessageTag.CHANNEL_NAME: channelName,
      MessageTag.STRING_CHANNEL_VALUE: value,
    };
    ReceivePort rport = ReceivePort();
    _sendPort.send([map, rport.sendPort]);
  }

  Future<String> getStringChannel(String channelName) async {
    //First make getControlChannel work, then implement this one
    String s = "";
    return s;
  }

  void setStringChannelCallback(String channelName, Function callback)
  {
    _callbackList.add(callback);
    Map<String, String> map = {
      MessageTag.MESSAGE_TYPE: MessageType.SET_STRING_CHANNEL_CALLBACK,
      MessageTag.CHANNEL_NAME: channelName,
      MessageTag.CALLBACK_INDEX: _callbackList.length.toString(),
    };
    _sendPort.send([map, _callbackPort.sendPort]);
  }

  void setTable(int table, List<double> value)
  {
    Map<String, String> map = {
      MessageTag.MESSAGE_TYPE: MessageType.SET_TABLE,
      MessageTag.TABLE_INDEX: table.toString(),
    };
    _sendPort.send([map, _callbackPort.sendPort, value]);
  }

  void setTableCallback(int table, Function callback)
  {
    _callbackList.add(callback);
    Map<String, String> map = {
      MessageTag.MESSAGE_TYPE: MessageType.SET_TABLE_CALLBACK,
      MessageTag.TABLE_INDEX: table.toString(), 
      MessageTag.CALLBACK_INDEX: _callbackList.length.toString(),
    };
    _sendPort.send([map, _callbackPort.sendPort]);
  }



  @override
  Future<int> getKsmps() async {
    Map<String, String> m = {MessageTag.MESSAGE_TYPE: MessageType.GET_KSMPS};
    ReceivePort rport = new ReceivePort();
    _sendPort.send([m, rport.sendPort]);

    int res = 0;
    await for(var msg in rport) {
      print(msg);
      res = int.parse(msg[0][MessageTag.MESSAGE_CONTENT]);
      break;
    }
    rport.close();
    return res;

  }

  @override
  Future<int> getNchnls()  async {
    Map<String, String> m = {MessageTag.MESSAGE_TYPE: MessageType.GET_NCHNLS};
    ReceivePort rport = new ReceivePort();
    _sendPort.send([m, rport.sendPort]);

    int res = 0;
    await for(var msg in rport) {
      print(msg);
      res = int.parse(msg[0][MessageTag.MESSAGE_CONTENT]);
      break;
    }
    rport.close();
    return res;
  }

  @override
  Future<int> getSr() async {
    Map<String, String> m = {MessageTag.MESSAGE_TYPE: MessageType.GET_SR};
    ReceivePort rport = new ReceivePort();
    _sendPort.send([m, rport.sendPort]);

    int res = 0;
    await for(var msg in rport) {
      print(msg);
      res = int.parse(msg[0][MessageTag.MESSAGE_CONTENT]);
      break;
    }
    rport.close();
    return res;
  }


  @override
  Future<double> evalCode(String code) async {
    Map<String, String> m = {MessageTag.MESSAGE_TYPE: MessageType.EVAL_CODE};
    ReceivePort rport = new ReceivePort();
    _sendPort.send([m, rport.sendPort, code]);
    double res = 0;
    await for(var msg in rport) {
      res = double.parse(msg[0][MessageTag.MESSAGE_CONTENT]);
      break;
    }
    rport.close();
    return res;
  }

  // ISOLATE METHODS (Mostly private threaded implementation)
  void _sendMessage(Map<String, String> msg)
  {
    //_isolateChannel.sink.add(msg);
    ReceivePort rport = ReceivePort();
    _sendPort.send([msg, rport.sendPort]);
  }

  void _startIsolate() async
  {
    _receivePort = ReceivePort();
    _isolate = await Isolate.spawn(CsoundNative._isolateEntryPoint, _receivePort.sendPort);
    _sendPort  = await _receivePort.first;
    //Callback system init
    _callbackList = [];
    _callbackPort = ReceivePort();
      this._initCallbackPort();
  }

  static Future<void> _perfAsync(Pointer<Void> csound,  
  Map<String, List<int>> controlTasks, Map<String, List<int>> stringTasks, Map<String, List<int>> tableTasks, Map<String, List<int>> audioTasks, Map<String,
          ConcurrentBuffer> audioChannels, SendPort callbackPort, Boolean keepPlaying) async
  {
    int ksmps = csoundGetKsmps(csound);

    keepPlaying.value = true;
    while(csoundPerformKsmps(csound) == 0 && keepPlaying.value == true) {
      await Future(() {});
      controlTasks.forEach((String key,List<int> value) {
        double val = csoundGetControlChannel(csound, StringUtf8Pointer(key).toNativeUtf8(), nullptr);
        value.forEach((int element) { // each element is index of the correspondant callback
           callbackPort.send([element, MessageTag.TYPE_CONTROL, val]); // send index, and value
        });
      });
      stringTasks.forEach((String key, List<int> value) {
        String val = csoundGetStringChannel(csound, key);
        value.forEach((int element) {
          callbackPort.send([element, MessageTag.TYPE_CONTROL, val]);
        });
      });
      tableTasks.forEach((key, value) {
        List<double> val = csoundGetTable(csound, int.parse(key));
        value.forEach((element) {
          callbackPort.send([element, MessageTag.TYPE_CONTROL, val]);
        });
      });

      audioTasks.forEach((String key,List<int> value) {
        csoundGetAudioChannel(csound, StringUtf8Pointer(key).toNativeUtf8(), audioChannels[key]!.getBuffer());
        audioChannels[key]!.unlock();
        value.forEach((int element) {
          callbackPort.send([element, MessageTag.TYPE_AUDIO, audioChannels[key]!.cBuf.address]);
        });
        // To do --> add a layer for csoundGetAudioChannel (convert ptr to List<double>)
      });
    }

    // Free memory
    audioChannels.forEach((key, value) {
      value.destroy();
    });

  }

  static void _isolateEntryPoint(SendPort sendPort) async
  {
    // functions index in main isolate
    Map<String, List<int>> isolateAudioChannelTasks = Map<String, List<int>>();
    Map<String, Pointer<Double>> audioChannelData = {};
    Map<String, ConcurrentBuffer> audioChannelBuffers = {};

    Map<String, List<int>> isolateControlChannelTasks = Map<String, List<int>>();
    Map<String, List<int>> isolateStringChannelTasks = Map<String, List<int>>(); 
    Map<String, List<int>> isolateTableTasks = Map<String, List<int>>(); 
    Boolean keepPlaying = Boolean(false);

    //print("Csound create ");
    Pointer<Void> csound = csoundCreate(nullptr);

    //print("Csound created");
    ReceivePort receivePort = ReceivePort();
    sendPort.send(receivePort.sendPort);
    SendPort callbackPort = receivePort.sendPort; // just for init to avoid null

    receivePort.listen((msg) {
      Map<String, String> map = msg[0];
      SendPort reply = msg[1];

      String type = map[MessageTag.MESSAGE_TYPE]!;

      if(type == MessageType.COMPILE_CSD_TEXT) {
        //print("Csound Compiling");
        csoundCompileCsdText(csound, StringUtf8Pointer(map[MessageTag.MESSAGE_CONTENT]!).toNativeUtf8());

        //reply.send("Csound Compiled");

      } else if(type == MessageType.COMPILE_CSD) {
        //csoundCompile(csound, Utf8.toUtf8(map[MessageTag.MESSAGE_CONTENT]));
        csoundCompileCsdText(csound, StringUtf8Pointer(map[MessageTag.MESSAGE_CONTENT]!).toNativeUtf8());

      } else if(type == MessageType.COMPILE_ORC) {
        csoundCompileOrc(csound, StringUtf8Pointer(map[MessageTag.MESSAGE_CONTENT]!).toNativeUtf8());

      } else if(type == MessageType.READ_SCORE) {
        csoundReadScore(csound, StringUtf8Pointer(map[MessageTag.MESSAGE_CONTENT]!).toNativeUtf8());

      } else if(type == MessageType.PERFORM) {
        csoundStart(csound);
        CsoundNative._perfAsync(csound,
         isolateControlChannelTasks, isolateStringChannelTasks, isolateTableTasks, isolateAudioChannelTasks, audioChannelBuffers, callbackPort, keepPlaying);
      
      } else if(type == MessageType.STOP) {
        keepPlaying.value = false;
        csoundStop(csound);
        //print("Stopping csound ");

      } else if(type == MessageType.SET_CONTROL_CHANNEL) {
        //print("CsoundSetControlChannel with ${double.parse(map[MessageTag.CONTROL_CHANNEL_VALUE]!)}");
        csoundSetControlChannel(csound, 
        StringUtf8Pointer(map[MessageTag.CHANNEL_NAME]!).toNativeUtf8(), double.parse(map[MessageTag.CONTROL_CHANNEL_VALUE]!));
        //reply.send("Control channel set");

      } else if(type == MessageType.GET_CONTROL_CHANNEL) {
        double val = csoundGetControlChannel(csound,
            StringUtf8Pointer(map[MessageTag.CONTROL_CHANNEL_VALUE]!)
                .toNativeUtf8(), nullptr);
        Map<String, String> m = Map<String, String>();
        m[MessageTag.MESSAGE_TYPE] = MessageType.GET_CONTROL_CHANNEL;
        m[MessageTag.CONTROL_CHANNEL_VALUE] = val.toString();
        sendPort.send(m);
      } else if(type == MessageType.GET_KSMPS) {
        int ksmps = csoundGetKsmps(csound);
        Map<String, String> m = {
          MessageTag.MESSAGE_TYPE: MessageType.GET_KSMPS,
          MessageTag.MESSAGE_CONTENT: ksmps.toString()
        };
        sendPort.send(m);
      } else if(type == MessageType.GET_NCHNLS) {
        int nchnls = csoundGetNchnls(csound);
        Map<String, String> m = {
          MessageTag.MESSAGE_TYPE: MessageType.GET_KSMPS,
          MessageTag.MESSAGE_CONTENT: nchnls.toString()
        };
        sendPort.send(m);
      } else if(type == MessageType.GET_SR) {
        int sr = csoundGetSr(csound).toInt();
        Map<String, String> m = {
          MessageTag.MESSAGE_TYPE: MessageType.GET_KSMPS,
          MessageTag.MESSAGE_CONTENT: sr.toString()
        };
        sendPort.send(m);
      } else if (type == MessageType.DELETE) {
        csoundDestroy(csound);

      } else if (type == MessageType.SET_CONTROL_CHANNEL_CALLBACK) {
        if (callbackPort == null) callbackPort = reply;
        if (!isolateControlChannelTasks.containsKey(
            map[MessageTag.CHANNEL_NAME]))
          isolateControlChannelTasks[map[MessageTag.CHANNEL_NAME]!] = [];
        isolateControlChannelTasks[map[MessageTag.CHANNEL_NAME]!]!.add(
            int.parse(map[MessageTag.CALLBACK_INDEX]!));


      } else if(type == MessageType.SET_AUDIO_CHANNEL_CALLBACK) {
        if (callbackPort == null) callbackPort = reply;
        if (!isolateAudioChannelTasks.containsKey(map[MessageTag.CHANNEL_NAME]))
          isolateAudioChannelTasks[map[MessageTag.CHANNEL_NAME]!] = [];

        int ksmps = csoundGetKsmps(csound);
        //print("ksmps = $ksmps" );
        isolateAudioChannelTasks[map[MessageTag.CHANNEL_NAME]!]!.add(
            int.parse(map[MessageTag.CALLBACK_INDEX]!));
        //audioChannelData[map[MessageTag.CHANNEL_NAME]!] = calloc.allocate(sizeOf<Double>() * ksmps);
        audioChannelBuffers[map[MessageTag.CHANNEL_NAME]!] = ConcurrentBuffer(ksmps);


      } else if(type == MessageType.SET_STRING_CHANNEL_CALLBACK) {
        if(callbackPort == null) callbackPort = reply;
        if(!isolateStringChannelTasks.containsKey(map[MessageTag.CHANNEL_NAME]))
         isolateStringChannelTasks[map[MessageTag.CHANNEL_NAME]!] = [];
        isolateStringChannelTasks[map[MessageTag.CHANNEL_NAME]!]!.add(int.parse(map[MessageTag.CALLBACK_INDEX]!));

      } else if(type == MessageType.INIT_CALLBACK_PORT) {
        callbackPort = reply;

      } else if(type == MessageType.SET_TABLE) {
        List<double> l = msg[2];
        csoundSetTable(csound, l, int.parse(map[MessageTag.TABLE_INDEX]!));

      } else if(type == MessageType.SET_TABLE_CALLBACK) {
        if(callbackPort == null) callbackPort = reply;
        if(!isolateTableTasks.containsKey(map[MessageTag.TABLE_INDEX]))
         isolateTableTasks[map[MessageTag.TABLE_INDEX]!] = [];
        isolateTableTasks[map[MessageTag.TABLE_INDEX]!]!.add(int.parse(map[MessageTag.CALLBACK_INDEX]!));
      } else if(type == MessageType.EVAL_CODE) {
        double res = csoundEvalCode(csound, StringUtf8Pointer(msg[2]).toNativeUtf8());
        Map<String, String> m = {
          MessageTag.MESSAGE_TYPE: MessageType.GET_KSMPS,
          MessageTag.MESSAGE_CONTENT: res.toString()
        };
        sendPort.send(m);
      }
    });
  }


  ReceivePort _callbackPort = ReceivePort();
  late List<Function> _callbackList;
  late Isolate _isolate;
  ReceivePort _receivePort = ReceivePort();
  late SendPort _sendPort;


}

Csound getCsoundDart()  {
  if(Platform.isAndroid) {
    return CsoundAndroid();
  }
  //else if(Platform.isIOS) {
  //}
  return CsoundNative();
}

Function getAppStartup() {
  return () {
    requestPermissions();
  };
}
